

class EntityError(Exception):
    def __init__(self, *args, message="", **kwargs):
        super().__init__(*args, **kwargs)
        self.nr = len(args)
        self.message = message

    def __repr__(self):
        result = f"<{type(self).__name__} ("
        if self.message != "":
            result += f"{self.message}: "
        for arg in self.args:
            result += f"{arg}, "
        result += ")>"
        return result

    def __eq__(self, other):
        if len(self.args) != len(other.args):
            return False
        for arg in self.args:
            if arg not in self.args:
                return False
        return True


class IllegalEntity(EntityError):
    pass


class IllegalEntities(EntityError):
    def __repr__(self):
        if self.message != "":
            message = f"{self.message}\n"
        else:
            message = ""
        result = f"IllegalEntities ({message}Number of errors: {self.nr},"
        for exception in self.args:
            result += f"\n    {repr(exception)}"
        result += ")"
        return result


class IllegalConfiguration(EntityError):
    pass


class IllegalConfigurations(EntityError):
    def __repr__(self):
        if self.message != "":
            message = f"{self.message}\n"
        else:
            message = ""
        result = f"IllegalConfigurations ({message}Number of errors: {self.nr},"
        for exception in self.args:
            result += f"\n    {repr(exception)}"
        result += ")"
        return result



from functools import reduce
from fractions import Fraction
from operator import mul


class Direction:
    up = 0
    right = 2
    down = 4
    left = 6

    def to_delta(direction):
        direction = direction % 8
        if direction == Direction.up:
            return (0, -1)
        elif direction == Direction.right:
            return (1, 0)
        elif direction == Direction.down:
            return (0, 1)
        elif direction == Direction.left:
            return (-1, 0)
        return (0, 0)


def catch(func, *args, handle=lambda e: e, exceptions=(Exception,), **kwargs):
    try:
        return func(*args, **kwargs)
    except exceptions as e:
        return handle(e)


def get_nr_of_permutations(nr_inputs, nr_outputs, max_nr):
    def nCk(n, k):
        return int(reduce(mul, (Fraction(n - i, i + 1) for i in range(k)), 1))

    perms = 0
    if nr_inputs < max_nr:
        max_nr = nr_inputs
    if nr_outputs < max_nr:
        max_nr = nr_outputs
    for i in range(1, max_nr + 1):
        perms += nCk(nr_inputs, i) * nCk(nr_outputs, i)
    return perms



from enum import Enum, auto
from fractions import Fraction


class Belt:
    def __init__(self, capacity=1, next=None, node=None):
        self.next = next
        self.node = node
        self.capacity = capacity
        self.__content = Fraction(0, 1)

    def __repr__(self):
        return f"<Belt( {self.__content} of {self.capacity} )>"

    @property
    def content(self):
        return self.__content

    @content.setter
    def content(self, value):
        if value > self.capacity:
            raise ValueError(
                f"Content can't exceed capacity: {value}/{self.capacity}")
        self.__content = value

    @property
    def available(self):
        return self.capacity - self.content

    @property
    def percentage(self):
        return (self.content / self.capacity) * 100.0

    @property
    def full(self):
        return self.content == self.capacity

    @property
    def empty(self):
        return self.content == 0

    def supply(self, amount=None):
        if amount is None or amount > self.capacity:
            amount = self.capacity
        if not isinstance(amount, Fraction):
            amount = Fraction(amount)
        result = amount - self.content
        self.content = amount
        return result

    def clear(self, amount=None):
        if amount is None:
            amount = self.content
        self.content -= amount
        return amount

    def transfer(self):
        if self.next is None:
            return

        available = self.next.available
        if available < self.content:
            self.content -= available
            self.next.content += available
        else:
            self.next.content += self.content
            self.content = 0

class Splitter:
    class Priority(Enum):
        off = None
        left = 'left'
        right = 'right'

    def __init__(self, entity=None,
                 uid=None,
                 lane_side=None, sideload=None, straight=None,
                 input_priority=None, output_priority=None):
        self.uid = uid
        self.entity = entity
        self.lane_side = lane_side
        self.sideload = sideload
        if self.entity is not None:
            self.entity._has_node = True
            if self.lane_side is not None:
                setattr(self.entity, f'_node_{self.lane_side}', self)
            elif self.sideload is not None:
                setattr(self.entity, f'_node_sideload_{self.sideload}', self)
            else:
                self.entity._node = self

        self.input_left = None
        self.input_right = None

        self.output_left = None
        self.output_right = None

        self._input_priority = input_priority
        self._output_priority = output_priority

    @property
    def input_priority(self):
        return self._input_priority or getattr(
            self.entity, 'input_priority', None)

    @property
    def output_priority(self):
        return self._output_priority or getattr(
            self.entity, 'output_priority', None)

    def get_inputs(self):
        inputs = [self.input_left, self.input_right]
        return [input for input in inputs if input]

    def get_outputs(self):
        outputs = [self.output_left, self.output_right]
        return [output for output in outputs if output]

    def get_available_inputs(self):
        inputs = [self.input_left, self.input_right]
        if self.input_priority == self.Priority.left.value:
            if self.input_left and self.input_left.content > 0:
                inputs = [self.input_left]
        elif self.input_priority == self.Priority.right.value:
            if self.input_right and self.input_right.content > 0:
                inputs = [self.input_right]
        inputs = [input for input in inputs
                  if input and input.content > 0]
        return inputs

    def get_available_outputs(self):
        outputs = [self.output_left, self.output_right]
        if self.output_priority == self.Priority.left.value:
            if self.output_left and self.output_left.available > 0:
                outputs = [self.output_left]
        elif self.output_priority == self.Priority.right.value:
            if self.output_right and self.output_right.available > 0:
                outputs = [self.output_right]
        outputs = [output for output in outputs
                   if output and output.available > 0]
        return outputs

    def balance(self):
        inputs = self.get_available_inputs()
        outputs = self.get_available_outputs()
        while inputs and outputs:
            available_space = min(
                [output.available for output in outputs]) * len(outputs)
            available_content = min(
                [input.content for input in inputs]) * len(inputs)
            if available_content > available_space:
                amount = available_space
            else:
                amount = available_content
            for input in inputs:
                input.content -= amount / len(inputs)
            for output in outputs:
                output.content += amount / len(outputs)
            inputs = self.get_available_inputs()
            outputs = self.get_available_outputs()

    @property
    def percentage(self):
        inputs = self.get_inputs()
        total = sum([input.available for input in inputs])
        return total / len(inputs)

    def make_networkx_graph(self, g, parent_name, node_colors, edge_colors):
        outputs = self.get_outputs()
        node_name = f'{self.uid}'
        if parent_name:
            edge_colors[(parent_name, node_name)] = self.percentage / 100.0
            g.add_edges_from([(parent_name, node_name)])
        for belt in outputs:
            if belt.next:
                belt.next.node.make_networkx_graph(
                    g, node_name, node_colors, edge_colors)
            elif belt in self.entity.blueprint._output_belts:
                for i, output in enumerate(self.entity.blueprint._output_belts):
                    if belt == output:
                        edge_colors[(node_name, f'output {i}')] = self.percentage / 100.0
                        g.add_edges_from([(node_name, f'output {i}')])
                        break



from py_factorio_blueprints.entity_mixins import BaseMixin
from py_factorio_blueprints.util import Vector
from factorio_balancers.exceptions import *
from factorio_balancers.graph import Belt as GraphBelt
from enum import Enum


class Connection:
    class Type(Enum):
        INPUT = 'input'
        OUTPUT = 'output'

    def __init__(self, entity, conn_type, multi_input=True):
        assert conn_type in Connection.Type
        self.__multi_input = multi_input
        self.__entity = entity
        self.__type = conn_type
        if self.type == Connection.Type.INPUT:
            self.__connections = []
        else:
            self.__connection = None

    @property
    def type(self):
        return self.__type

    @property
    def entity(self):
        if not (self.type == Connection.Type.OUTPUT or not self.__multi_input):
            raise AttributeError("Connection has no attribute 'entity'")
        if self.__connection is None:
            return None
        return self.__connection.__entity

    @property
    def connected(self):
        if self.type == Connection.Type.INPUT:
            return len(self.__connections) > 0
        else:
            return self.__connection is not None

    @property
    def entities(self):
        if not (self.type == Connection.Type.INPUT and self.__multi_input):
            raise AttributeError("Connection has no attribute 'entities'")
        return [conn.__entity for conn in self.__connections]

    def connect(self, other):
        if not isinstance(other, Connection):
            raise TypeError(Connection)
        if self.type == Connection.Type.INPUT:
            assert other.type == Connection.Type.OUTPUT
            other.__connection = self
            if not self.__multi_input:
                assert len(self.__connections) == 0
            assert len(self.__connections) < 3
            if other not in self.__connections:
                self.__connections.append(other)
        else:
            assert other.type == Connection.Type.INPUT
            self.__connection = other
            assert len(other.__connections) < 3
            if self not in other.__connections:
                other.__connections.append(self)

    def disconnect(self, other):
        assert isinstance(other, Connection)
        if self.type == Connection.Type.INPUT:
            assert other.type == Connection.Type.OUTPUT
            self.__connections.remove(other)
            other.__connection = None
        else:
            assert other.type == Connection.Type.INPUT
            self.__connection = None
            other.__connections.remove(self)


class BalancerEntity(BaseMixin):
    def input_belt_check(self):
        if not self.forward.connected:
            raise IllegalConfiguration(
                self,
                message="Entity has no forward partner")
        elif self.forward.entity.has_sideloads:
            return False
        else:
            return self.forward.entity.input_belt_check()

    def side_from_position(self, position):
        rot_amount = self.direction // -2
        side_vector = position - self.position
        side_vector = side_vector.rotate(rot_amount)
        if side_vector.x < 0:
            return 'left'
        elif side_vector.x > 0:
            return 'right'
        else:
            return None

    @property
    def coordinates(self):
        return [self.position]

    def pad_connection(self, inp=False, out=False):
        names = {
            1: 'transport-belt',
            2: 'fast-transport-belt',
            3: 'express-transport-belt',
        }
        name = names[self.name.data['belt_speed']]

        if inp:
            new = self.blueprint.entities.make(
                name=name,
                position=self.position + self.direction.rotate(2).vector,
                direction=self.direction)
            connection1 = self.backward
            connection2 = new.forward
            connection1.connect(connection2)
        if out:
            new = self.blueprint.entities.make(
                name=name,
                position=self.position + self.direction.vector,
                direction=self.direction)
            connection1 = self.forward
            connection2 = new.backward
            connection1.connect(connection2)

    def _trace_nodes(self, belt, position, lane=None):
        speed = self.name.data['belt_speed']
        if speed < belt.capacity:
            belt.capacity = speed
        if isinstance(self.forward.entity, (Belt, Underground)):
            return self.forward.entity._trace_nodes(belt, self.position, lane=lane)
        elif isinstance(self.forward.entity, Splitter):
            self.forward.entity._trace_nodes(belt, self.position, lane=lane)
            return belt
        elif self.forward.entity is None:
            self.blueprint._belts.append(belt)
            self.blueprint._output_belts.append(belt)


class Belt(BalancerEntity):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__forward = Connection(self, Connection.Type.OUTPUT)
        self.__backward = Connection(self, Connection.Type.INPUT)

    def _trace_nodes(self, belt, position, lane=None):
        if lane is None or not getattr(self, '_has_node', False):
            return super()._trace_nodes(belt, self.position, lane=lane)

        def opposite(a):
            return {'left': 'right', 'right': 'left'}[a]

        speed = self.name.data['belt_speed']
        if speed < belt.capacity:
            belt.capacity = speed
        side = self.side_from_position(position)
        if side is None:
            node = getattr(self, f'_node_{lane}')
            field = f'input_{opposite(lane)}'
            if getattr(node, field) is None:
                setattr(node, field, GraphBelt(capacity=speed, node=node))
                belt.next = getattr(node, field)
            else:
                raise IllegalConfiguration(
                    "Can't connect to the same side twice")
        else:
            node = getattr(self, f'_node_sideload_{side}')
            field = f'input_{lane}'
            if getattr(node, field) is None:
                setattr(node, field, GraphBelt(capacity=speed, node=node))
                belt.next = getattr(node, field)
            else:
                raise IllegalConfiguration(
                    "Can't connect to the same side twice")
        self.blueprint._belts.append(belt)

        lane_side = lane if side is None else side
        if getattr(self, f'_traversed_{lane_side}', False):
            return belt
        setattr(self, f'_traversed_{lane_side}', True)

        node = getattr(self, f'_node_{lane_side}')
        field = f'output_{opposite(lane_side)}'
        setattr(node, field, GraphBelt(capacity=speed, node=node))
        super()._trace_nodes(
            getattr(node, field),
            self.position, lane=lane_side)
        return belt

    @property
    def forward(self):
        return self.__forward

    @property
    def backward(self):
        return self.__backward

    def reset(self):
        pass

    def get_connection_for(self, position, conn_type):
        if conn_type == Connection.Type.INPUT:
            return self.backward
        else:
            return self.forward

    @property
    def has_sideloads(self):
        return len(self.backward.entities) > 1

    @property
    def has_no_inputs(self):
        return not self.backward.connected

    @property
    def has_no_outputs(self):
        return not self.forward.connected

    def setup_transport_lines(self):
        position = self.position + self.direction.vector
        other = self.blueprint.entities[position]
        if not other:
            return
        elif len(other) > 1:
            raise IllegalConfiguration(
                self, other,
                message="More than one entity occupying the same space")
        other = other[0]
        if other.direction == self.direction.rotate(2):
            raise IllegalConfiguration(
                self, other, message="Entities facing each other")
        if other.direction != self.direction and isinstance(other, Splitter):
            raise IllegalConfiguration(
                self, other, message="Sideloading onto a splitter")
        self.forward.connect(
            other.get_connection_for(
                position, Connection.Type.INPUT))


class Splitter(BalancerEntity):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__forward_left = Connection(
            self, Connection.Type.OUTPUT, multi_input=False)
        self.__forward_right = Connection(
            self, Connection.Type.OUTPUT, multi_input=False)
        self.__backward_left = Connection(self, Connection.Type.INPUT)
        self.__backward_right = Connection(self, Connection.Type.INPUT)

    @property
    def forward_left(self):
        return self.__forward_left

    @property
    def forward_right(self):
        return self.__forward_right

    @property
    def backward_left(self):
        return self.__backward_left

    @property
    def backward_right(self):
        return self.__backward_right

    def reset(self):
        pass

    @property
    def coordinates(self):
        amount = self.direction // 2
        vectors = [
            Vector(-0.5, 0).rotate(amount),
            Vector(0.5, 0).rotate(amount)]
        return [self.position + vector
                for vector in vectors]

    def get_connection_for(self, position, conn_type):
        rot_amount = self.direction // -2

        side_vector = position - self.position
        side_vector = side_vector.rotate(rot_amount)
        if side_vector.x < 0:
            if conn_type == Connection.Type.INPUT:
                return self.backward_left
            else:
                return self.forward_left
        else:
            if conn_type == Connection.Type.INPUT:
                return self.backward_right
            else:
                return self.forward_right

    @property
    def has_sideloads(self):
        return False

    def input_belt_check(self):
        return True

    @property
    def has_no_inputs(self):
        return not self.backward_right.connected and \
               not self.backward_left.connected

    @property
    def has_no_outputs(self):
        return not self.forward_left.connected and \
               not self.forward_right.connected

    def setup_transport_lines(self):
        positions = [coordinate + self.direction.vector
                     for coordinate in self.coordinates]
        for position in positions:
            other = self.blueprint.entities[position]
            if not other:
                continue
            elif len(other) > 1:
                raise IllegalConfiguration(
                    self, other,
                    message="More than one entity occupying the same space")
            other = other[0]
            if other.direction == self.direction.rotate(2):
                raise IllegalConfiguration(
                    self, other, message="Entities facing each other")
            if other.direction != self.direction and isinstance(other, Splitter):
                raise IllegalConfiguration(
                    self, other, message="Sideloading onto a splitter")
            connection1 = self.get_connection_for(
                position, Connection.Type.OUTPUT)
            connection2 = other.get_connection_for(
                position, Connection.Type.INPUT)
            connection1.connect(connection2)

    def pad_connection(self, inp=False, out=False):
        names = {
            1: 'transport-belt',
            2: 'fast-transport-belt',
            3: 'express-transport-belt',
        }
        name = names[self.name.data['belt_speed']]

        left, right = self.coordinates
        if inp:
            new_left = self.blueprint.entities.make(
                name=name,
                position=left + self.direction.rotate(2).vector,
                direction=self.direction)
            new_right = self.blueprint.entities.make(
                name=name,
                position=right + self.direction.rotate(2).vector,
                direction=self.direction)
            self.backward_left.connect(new_left.forward)
            self.backward_right.connect(new_right.forward)
        if out:
            new_left = self.blueprint.entities.make(
                name=name,
                position=left + self.direction.vector,
                direction=self.direction)
            new_right = self.blueprint.entities.make(
                name=name,
                position=right + self.direction.vector,
                direction=self.direction)
            self.forward_left.connect(new_left.backward)
            self.forward_right.connect(new_right.backward)

    def _trace_nodes(self, belt, position, lane=None):
        side = self.side_from_position(position)
        node = getattr(self, f'_node{f"_{lane}" if lane is not None else ""}')
        if side == 'left':
            if node.input_left is None:
                node.input_left = GraphBelt(
                    capacity=belt.capacity,
                    node=node)
                belt.next = node.input_left
            else:
                raise IllegalConfiguration(
                    "Can't connect to the same side twice")
        else:
            if node.input_right is None:
                node.input_right = GraphBelt(
                    capacity=belt.capacity,
                    node=node)
                belt.next = node.input_right
            else:
                raise IllegalConfiguration(
                    "Can't connect to the same side twice")
        self.blueprint._belts.append(belt)

        if getattr(self, f'_traversed{f"_{lane}" if lane else ""}', False):
            return belt
        setattr(self, f'_traversed{f"_{lane}" if lane else ""}', True)

        if self.forward_left.connected:
            node.output_left = GraphBelt(
                capacity=self.name.data['belt_speed'],
                node=node)
            self.forward_left.entity._trace_nodes(
                node.output_left, self.coordinates[0], lane=lane)
        if self.forward_right.connected:
            node.output_right = GraphBelt(
                capacity=self.name.data['belt_speed'],
                node=node)
            self.forward_right.entity._trace_nodes(
                node.output_right, self.coordinates[1], lane=lane)

        return belt


class Underground(BalancerEntity):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__forward = Connection(self, Connection.Type.OUTPUT)
        self.__backward = Connection(self, Connection.Type.INPUT)

    @property
    def forward(self):
        return self.__forward

    @property
    def backward(self):
        return self.__backward

    def reset(self):
        pass

    @property
    def has_no_inputs(self):
        return not self.backward.connected
    @property

    def has_no_outputs(self):
        return not self.forward.connected

    def get_connection_for(self, position, conn_type):
        if conn_type == Connection.Type.INPUT:
            return self.backward
        else:
            return self.forward

    def find_partner(self):
        max_distance = self.name.data['max_distance'] + 1
        if self.type == 'output':
            direction = self.direction.rotate(2)
        else:
            direction = self.direction
        vector = direction.vector
        for i in range(1, max_distance + 1):
            entity = self.blueprint.entities[self.position + vector * i]
            if not entity:
                continue
            elif len(entity) > 1:
                raise IllegalConfiguration(
                    self, entity,
                    message="More than one entity occupying the same space")
            entity = entity[0]
            if entity.name != self.name:
                continue
            elif entity.direction == self.direction \
                    and self.type == 'input' \
                    and entity.type == 'input':
                raise IllegalConfiguration(
                    self, entity, message="Two underground inputs in a row")
            elif entity.direction == self.direction.rotate(2) \
                    and self.type == 'input':
                raise IllegalConfiguration(
                    self, entity,
                    message="Two connected underground inputs")
            elif entity.direction == self.direction.rotate(2) \
                    and self.type == 'output' \
                    and entity.type == 'output':
                raise IllegalConfiguration(
                    self, entity,
                    message="Two connected underground outputs")
            elif entity.direction == self.direction \
                    and entity.type != self.type:
                return entity
        return None

    @property
    def has_sideloads(self):
        for entity in self.backward.entities:
            if entity.direction != self.direction:
                return True
        else:
            return False

    def setup_transport_lines(self):
        partner = self.find_partner()
        if partner is None and self.type == 'input':
            raise IllegalConfiguration(
                self, message="Lone underground input")
        if partner is not None:
            if self.type == 'input':
                self.forward.connect(
                    partner.get_connection_for(
                        partner.position, Connection.Type.INPUT))
            else:
                self.backward.connect(
                    partner.get_connection_for(
                        partner.position, Connection.Type.OUTPUT))
        if self.type == 'output':
            Belt.setup_transport_lines(self)

    def _trace_nodes(self, belt, position, lane=None):
        if lane is None or not getattr(self, '_has_node', False):
            return super()._trace_nodes(belt, self.position, lane=lane)

        def opposite(a):
            return {'left': 'right', 'right': 'left'}[a]

        speed = self.name.data['belt_speed']
        if speed < belt.capacity:
            belt.capacity = speed
        side = self.side_from_position(position)
        if side is None:
            node = getattr(self, f'_node_{lane}')
            field = f'input_{opposite(lane)}'
            if getattr(node, field) is None:
                setattr(node, field, GraphBelt(capacity=speed, node=node))
                belt.next = getattr(node, field)
            else:
                raise IllegalConfiguration(
                    "Can't connect to the same side twice")
        else:
            node = getattr(self, f'_node_{side}')
            lane_check = lane
            if self.type == 'output':
                lane_check = opposite(lane)
            if side != lane_check:
                return belt
            field = f'input_{side}'
            if getattr(node, field) is None:
                setattr(node, field, GraphBelt(
                    capacity=speed, node=node))
                belt.next = getattr(node, field)
            else:
                raise IllegalConfiguration(
                    "Can't connect to the same side twice")
        self.blueprint._belts.append(belt)

        lane_side = lane if side is None else side
        if getattr(self, f'_traversed_{lane_side}', False):
            return belt
        setattr(self, f'_traversed_{lane_side}', True)

        node = getattr(self, f'_node_{lane_side}')
        field = f'output_{opposite(lane_side)}'
        setattr(node, field, GraphBelt(capacity=speed, node=node))
        super()._trace_nodes(
            getattr(node, field),
            self.position, lane=lane_side)
        return belt


entity_prototypes = {
    'transport-belt': {
        'mixins': [Belt],
    },
    'underground-belt': {
        'mixins': [Underground],
    },
    'splitter': {
        'mixins': [Splitter],
    },
}



from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='factorio_balancers',
    packages=['factorio_balancers'],
    version='0.2.3',
    license='MIT',
    description='A property testing simulator for balancers from the game Factorio',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Tijmen Zwaan',
    author_email='tijmen.zwaan@gmail.com',
    url='https://github.com/tzwaan/factorio_balancers',
    download_url='https://github.com/tzwaan/factorio_balancers/archive/v0.2.3.tar.gz',
    keywords=['factorio', 'balancer', 'blueprint'],
    install_requires=[
        'py_factorio_blueprints',
        'progress',
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Natural Language :: English'
    ],
    scripts=[
        'bin/balancer_test'
    ],
    include_package_data=True,
    zip_safe=False)



import logging
import os
from itertools import combinations
from progress.bar import Bar
from fractions import Fraction
from py_factorio_blueprints import Blueprint
from py_factorio_blueprints.util import Vector
from factorio_balancers.utils import catch, get_nr_of_permutations
from factorio_balancers.graph import Splitter, Belt
from factorio_balancers.entity_mixins import (
    entity_prototypes, Splitter as SplitterMixin,
    Belt as BeltMixin, Underground as UndergroundMixin)
from factorio_balancers.exceptions import *


logger = logging.getLogger("factorio_balancers.balancer")


class MyBar(Bar):
    def finish(self, clear=True):
        if clear:
            self.clearln()
            print('\x1b[?25h', end='')
        else:
            super().finish()

    @property
    def eta_display(self):
        return str(self.eta_td)


class OptionalBar:
    def __init__(self, *args, verbose, **kwargs):
        self.verbose = verbose
        if kwargs.get('suffix', None) is None:
            kwargs['suffix'] = '%(percent)d%% - %(eta_display)s'
        if self.verbose:
            self.bar = MyBar(*args, **kwargs)

    def finish(self, *args, **kwargs):
        if self.verbose:
            self.bar.finish(*args, **kwargs)

    def next(self, *args, **kwargs):
        if self.verbose:
            self.bar.next(*args, **kwargs)


def is_close(a, b, rel_tol=1e-06, abs_tol=0.0):
    return abs(a - b) <= max(rel_tol * max(abs(a), abs(b)), abs_tol)


class Balancer(Blueprint):
    def __init__(self, *args, **kwargs):
        Blueprint.import_prototype_data(
            f"{os.path.dirname(__file__)}/entity_data.json")
        super().__init__(
            *args, custom_entity_prototypes=entity_prototypes, **kwargs)

        self.recompile_entities()
        logger.debug("Before padding")
        logger.debug(self.print2d())
        self.pad_connections()
        logger.debug("After padding")
        logger.info(self.print2d())
        inputs, outputs = self._get_external_connections()
        self.nr_inputs = len(inputs)
        self.nr_outputs = len(outputs)
        logger.debug(f"Nr of inputs and outputs: "
                     f"{self.nr_inputs}, {self.nr_outputs}")
        logger.debug(f"Inputs: {inputs}")
        logger.debug(f"Outputs: {outputs}")

        nodes = self._get_nodes()
        logger.debug(f"Number of nodes {len(nodes)}")
        logger.debug(f"Nodes {nodes}")

        for i, node in enumerate(nodes):
            node.node_number = i

        self.generate_simulation()
        self.nr_inputs_sim = len(self._input_belts)
        self.nr_outputs_sim = len(self._output_belts)

    def print2d(self):
        maxx, minx, maxy, miny = self.maximum_values
        width = maxx - minx + 1
        height = maxy - miny + 1
        offset = Vector(minx, miny)
        data = [[" " for _ in range(width)] for _ in range(height)]

        for entity in self.entities:
            for i, coord in enumerate(entity.coordinates):
                position = coord - offset
                x, y = position.round()
                data[y][x] = entity.name.data['ascii'][entity.direction // 2][i]
        result = "    "
        for line in data:
            result += ''.join(line)
            result += '\n    '
        return f'\n{result}\n'

    def _get_nodes(self):
        return [
            entity
            for entity in self.entities
            if isinstance(entity, SplitterMixin) or entity.has_sideloads]

    def generate_simulation(self):
        self._splitters = []
        self._belts = []
        self._inputs = []
        self._outputs = []

        if self.has_sideloads:
            self._parse_lane_balancer()
            return
        else:
            self._parse_balancer()

        if self._verbose:
            logger.debug(f"Inputs: {self._input_belts}")
            logger.debug(f"Outputs: {self._output_belts}")
            logger.debug(f"Splitters: {self._splitters}")
        for entity in self._get_nodes():
            if not entity._traversed:
                raise IllegalConfiguration(
                    message="The balancer is not fully connected")

    def cycle(self):
        for splitter in self._splitters:
            splitter.balance()
        for belt in self._belts:
            belt.transfer()

    def clear(self):
        total = 0
        for belt in self._belts:
            total += belt.clear()
            if belt.next:
                total += belt.next.clear()
        return total

    def fill(self):
        total = 0
        for belt in self._belts:
            total += belt.supply()
            if belt.next:
                total += belt.next.supply()
        return total

    def supply(self, *args, **kwargs):
        return [
            belt.supply(*args, **kwargs)
            for belt in self._input_belts]

    def drain(self, **kwargs):
        return [
            belt.clear(**kwargs)
            for belt in self._output_belts]

    def recompile_entities(self):
        for entity in self.entities:
            entity.reset()
        exceptions = self.setup_transport_lines()
        nr_exceptions = len(exceptions)
        if nr_exceptions > 0:
            logger.info(self.print2d())
            raise IllegalConfigurations(*exceptions)

        self.has_sideloads = self._has_sideloads()

    def _get_external_connections(self):
        inputs = []
        outputs = []
        inputs = [entity
                  for entity in self.entities
                  if entity.has_no_inputs and entity.input_belt_check()]
        outputs = [entity
                   for entity in self.entities
                   if entity.has_no_outputs]
        return inputs, outputs

    def _input_belt_check(self, entity):
        if entity._is_splitter:
            return True
        elif entity._partner_forward is None:
            raise IllegalConfiguration(
                entity,
                message="Entity has no forward partner")
        elif entity._partner_forward.has_sideloads:
            return False
        else:
            return self._input_belt_check(entity._partner_forward)

    def pad_connections(self):
        inputs, outputs = self._get_external_connections()

        self.pad_entities(inputs, inp=True)
        self.pad_entities(outputs, out=True)

    def pad_entities(self, entities, **kwargs):
        for entity in entities:
            if isinstance(entity, SplitterMixin):
                break
        else:
            return
        for entity in entities:
            entity.pad_connection(**kwargs)

    def setup_transport_lines(self):
        exceptions = []
        for entity in self.entities:
            e = catch(entity.setup_transport_lines,
                      exceptions=IllegalConfiguration)
            if e is not None and e not in exceptions:
                exceptions.append(e)
        return exceptions

    def _has_sideloads(self):
        for entity in self.entities:
            if entity.has_sideloads:
                return True
        return False

    def _parse_balancer(self):
        for i, entity in enumerate(self._get_nodes()):
            splitter = Splitter(entity=entity, uid=i)
            self._splitters.append(splitter)
        inputs, outputs = self._get_external_connections()
        self._input_belts = []
        self._output_belts = []
        for input in inputs:
            assert isinstance(input, (BeltMixin, UndergroundMixin))
            belt = input._trace_nodes(
                Belt(capacity=input.name.data['belt_speed']),
                input.position)
            assert len(outputs) == len(self._output_belts)
            self._input_belts.append(belt)

    def _parse_lane_balancer(self):
        for entity in self._get_nodes():
            if isinstance(entity, SplitterMixin):
                splitter_left = Splitter(entity=entity, lane_side='left')
                splitter_right = Splitter(entity=entity, lane_side='right')
                self._splitters.append(splitter_left)
                self._splitters.append(splitter_right)
            elif isinstance(entity, BeltMixin):
                sideload_left = Splitter(
                    entity=entity, sideload='left',
                    input_priority=Splitter.Priority.right.value)
                straight_left = Splitter(
                    entity=entity, lane_side='left',
                    input_priority=Splitter.Priority.right.value)
                straight_left.input_left = Belt(
                    capacity=entity.name.data['belt_speed'],
                    node=straight_left)
                sideload_left.output_left = Belt(
                    capacity=entity.name.data['belt_speed'],
                    node=sideload_left,
                    next=straight_left.input_left)
                self._belts.append(sideload_left.output_left)
                sideload_right = Splitter(
                    entity=entity, sideload='right',
                    input_priority=Splitter.Priority.left.value)
                straight_right = Splitter(
                    entity=entity, lane_side='right',
                    input_priority=Splitter.Priority.left.value)
                straight_right.input_right = Belt(
                    capacity=entity.name.data['belt_speed'],
                    node=straight_right)
                sideload_right.output_right = Belt(
                    capacity=entity.name.data['belt_speed'],
                    node=sideload_right,
                    next=straight_right.input_right)
                self._belts.append(sideload_right.output_right)
                self._splitters.append(sideload_left)
                self._splitters.append(straight_left)
                self._splitters.append(sideload_right)
                self._splitters.append(straight_right)
            elif isinstance(entity, UndergroundMixin):
                sideload_left = Splitter(
                    entity=entity, lane_side='left',
                    input_priority=Splitter.Priority.right.value)
                sideload_right = Splitter(
                    entity=entity, lane_side='right',
                    input_priority=Splitter.Priority.left.value)
                self._splitters.append(sideload_left)
                self._splitters.append(sideload_right)
            else:
                raise Exception('what')
        inputs, outputs = self._get_external_connections()
        self._input_belts = []
        self._output_belts = []
        for input in inputs:
            assert isinstance(input, (BeltMixin, UndergroundMixin))
            belt_left = input._trace_nodes(
                Belt(capacity=input.name.data['belt_speed']),
                input.position,
                lane='left')
            belt_right = input._trace_nodes(
                Belt(capacity=input.name.data['belt_speed']),
                input.position,
                lane='right')
            assert len(outputs) * 2 == len(self._output_belts)
            self._input_belts.append(belt_left)
            self._input_belts.append(belt_right)
        for i, splitter in enumerate(self._splitters):
            splitter.uid = i

    def graph_check(self):
        """
        Recursively traverse all the nodes from a single input, going both
        forward and backwards through the graph.
        Checks whether all nodes have been traversed. If not, this means the
        graph is not fully connected.
        """
        logger.debug("All nodes:")
        for node in self._get_nodes():
            logger.debug(f"    {node}")
            node._traversed = False

        start_node = self._inputs[0]
        logger.debug(f"start node: {start_node}")

        self._traverse_node(start_node)
        for node in self.nodes:
            if not node._traversed:
                return False
        return True

    def _traverse_node(self, node):
        logger.debug(f"traversing node: {node}")
        if node is None or node._traversed:
            return
        node._traversed = True

        if type(node) is Belt:
            self._traverse_node(node._input)
            self._traverse_node(node._output)
        else:
            self._traverse_node(node._input_left)
            self._traverse_node(node._input_right)
            self._traverse_node(node._output_left)
            self._traverse_node(node._output_right)

    def make_networkx_graph(self):
        import networkx as nx
        import matplotlib.pyplot as plt
        g = nx.DiGraph()
        node_colors = {}
        edge_colors = {}
        for i, belt in enumerate(self._input_belts):
            node_name = f'input {i}'
            if belt.next:
                belt.next.node.make_networkx_graph(
                    g, node_name, node_colors, edge_colors)

        pos = nx.spring_layout(g)
        # nx.draw(g)
        # nx.draw_networkx_nodes(g, pos, node_size=500)
        nx.draw_networkx_labels(g, pos)
        edge_colors = [edge_colors.get(edge, 0.0) for edge in g.edges()]
        nx.draw(
            g, pos,
            edge_color=edge_colors,
            edge_cmap=plt.get_cmap('jet'),
            arrows=True)
        # nx.draw_networkx_edges(g, pos, arrows=True)
        filename = f"{os.path.dirname(__file__)}/../graph.png"
        plt.show()

    def test_output_balance(
            self, verbose=False, trickle=False, debug=False, **kwargs):
        bar = OptionalBar(
            '   -- Progress',
            verbose=verbose,
            max=len(self._input_belts) + 1)

        amount = Fraction(1, 4) if trickle else None
        for input in self._input_belts:
            self.clear()
            drained = self.drain()
            supplied = input.supply(amount)
            while not is_close(sum(drained), supplied):
                self.cycle()
                drained = self.drain()
                supplied = input.supply(amount)

            if len(set(drained)) > 1:
                bar.finish()
                return False
            bar.next()

        self.clear()
        drained = self.drain()
        supplied = self.supply(amount)
        while not is_close(sum(drained), sum(supplied)):
            self.cycle()
            drained = self.drain()
            supplied = self.supply(amount)
        bar.finish()
        if len(set(drained)) > 1:
            return False
        return True

    def test_input_balance(self, verbose=False, trickle=False, **kwargs):
        bar = OptionalBar(
            '   -- Progress',
            verbose=verbose,
            max=len(self._output_belts) + 1)

        amount = Fraction(1, 4) if trickle else None
        for output in self._output_belts:
            self.fill()
            drained = output.clear(amount)
            supplied = self.supply()
            while not is_close(drained, sum(supplied)):
                self.cycle()
                drained = output.clear(amount)
                supplied = self.supply()

            if len(set(supplied)) > 1:
                bar.finish()
                return False
            bar.next()

        self.fill()
        drained = self.drain(amount=amount)
        supplied = self.supply()
        while not is_close(sum(drained), sum(supplied)):
            self.cycle()
            drained = self.drain(amount=amount)
            supplied = self.supply()
        bar.finish()
        if len(set(supplied)) > 1:
            return False
        return True

    def test_throughput(
            self, inputs=None, outputs=None, verbose=False, **kwargs):
        if inputs is None:
            inputs = self._input_belts
        if outputs is None:
            outputs = self._output_belts

        def drain(a):
            return [b.clear() for b in a]

        def supply(a):
            return [b.supply() for b in a]

        self.clear()
        drained = drain(outputs)
        supplied = supply(inputs)
        clean_input = supplied
        while not is_close(sum(drained), sum(supplied)):
            drained = drain(outputs)
            self.cycle()
            supplied = supply(inputs)

        worst_percentage = 100.0
        for output in outputs:
            percentage = output.percentage
            if not is_close(worst_percentage, percentage) and \
                    percentage < worst_percentage:
                worst_percentage = percentage
        drained = drain(outputs)

        if is_close(sum(clean_input), sum(drained)):
            return True, worst_percentage
        if worst_percentage < 100.0:
            return False, worst_percentage
        return True, worst_percentage

    def throughput_sweep(self, extensive=False, verbose=False, **kwargs):
        if extensive:
            i_range = range(
                1,
                min(
                    len(self._input_belts),
                    len(self._output_belts)))
            nr_of_permutations = get_nr_of_permutations(
                len(self._input_belts),
                len(self._output_belts),
                len(self._input_belts))
        else:
            i_range = range(1, 3)
            nr_of_permutations = get_nr_of_permutations(
                len(self._input_belts),
                len(self._output_belts),
                2)
        bar = OptionalBar(
            '   -- Progress', verbose=verbose, max=nr_of_permutations)

        results = []

        for i in i_range:
            for inputs in combinations(self._input_belts, i):
                for outputs in combinations(self._output_belts, i):
                    results.append(
                        self.test_throughput(inputs=inputs, outputs=outputs))
                    bar.next()
        bar.finish()
        return results

    def test_throughput_unlimited(self, **kwargs):
        results = self.throughput_sweep(**kwargs)
        limited = False
        worst = 100.0
        for unlimited, percentage in results:
            if not unlimited:
                limited = True
                if not is_close(worst, percentage) and \
                        percentage < worst:
                    worst = percentage

        return not limited, worst

    def test(self, balance=True, throughput=True, trickle=False,
             sweep=False, extensive_sweep=False, verbose=False):
        self.clear()
        is_lane_balancer = self.has_sideloads
        logger.info(
            f"Testing a {len(self._input_belts)} - "
            f"{len(self._output_belts)} "
            f"{'lane ' if is_lane_balancer else ''}balancer.")

        results = {}

        if balance:
            logger.info("\n  Testing balance.")
            output_balanced = self.test_output_balance(verbose=verbose)
            results['output_balanced'] = output_balanced
            logger.info(
                f"   -- Output is {'' if output_balanced else 'NOT '}balanced.")

            input_balanced = self.test_input_balance(verbose=verbose)
            results['input_balanced'] = input_balanced
            logger.info(
                f"   -- Input is {'' if input_balanced else 'NOT '}balanced.")
        if trickle:
            logger.info("\n  Testing balance using trickle.")
            output_balanced = self.test_output_balance(
                verbose=verbose, trickle=True)
            results['output_balanced_trickle'] = output_balanced
            logger.info(
                f"   -- Output is {'' if output_balanced else 'NOT '}"
                f"balanced with a trickle.")
            input_balanced = self.test_input_balance(
                verbose=verbose, trickle=True)
            results['input_balanced_trickle'] = input_balanced
            logger.info(
                f"   -- Input is {'' if input_balanced else 'NOT '}"
                f"balanced with a trickle.")

        if throughput:
            full_throughput, worst = self.test_throughput(verbose=verbose)
            logger.info("\n  Testing regular throughput.")
            results['full_throughput'] = full_throughput
            if full_throughput:
                logger.info("   -- Full throughput on regular use")
            else:
                results['full_throughput_bottleneck'] = worst
                logger.info(
                    f"   -- Limited throughput to {worst} on "
                    f"regular use on at least one of the outputs.")

        if sweep or extensive_sweep:
            logger.info(
                f"\n  {'Extensive' if extensive_sweep else 'Regular'} "
                f"throughput sweep")
            unlimited, worst = self.test_throughput_unlimited(
                extensive=extensive_sweep, verbose=verbose)
            if not unlimited:
                logger.info(
                    f"   -- At least one bottleneck exists that "
                    f"limits throughput to {worst}%.")
                results['largest_bottleneck'] = worst
            else:
                logger.info(
                    f"   -- No bottlenecks with any combinations of"
                    f" {'any number of' if extensive_sweep else '1 or 2'} "
                    f"any number of inputs and outputs.")

            if extensive_sweep:
                results['throughput_unlimited'] = unlimited
            else:
                results['throughput_unlimited_candidate'] = unlimited

        logger.info("\n")
        return results
